package require nx
package require tdom
package provide ::TAGG 0.1

# Matrix-Reporter Component 
namespace eval TAGG {

  nx::Class create TraceAggregator {

    # basic elements
    :property -accessor public {matrix}
    :property -accessor public {allCases}
    :property -accessor public {allScenarios}
    :property -accessor public {allCalls}
    :property -accessor public {allFeatures}
    :property -accessor public {allInstances}
    :property -accessor public {allClasses}
    :property -accessor public {aggCalls}

    # build a test-to-class matrix (mapping test cases to SUT classes)
    # <xmiFile> (path to source XMI file)
    # <testLevel> (case/scenario)
    # <factor> (feature/call)
    # <featureType> (called/calling/interClassCalling)
    # <callScope> (all/testScript/sut)
    # <list> (0/1)
    :public method buildTestToSutMatrix { xmiFile testLevel factor featureType callScope list} {

      :matrix set [TestClassMatrix new]
      :setBasicElements $xmiFile

      foreach c [:allCases get] {
        set testCaseName [$c getAttribute name]

        # handling calls in case blocks
        set caseBlocks [$c selectNodes block]
        foreach b $caseBlocks {
          set blockCalls [$b getAttribute call]
          :aggCalls set [expr {[:aggCalls get] + [llength $blockCalls]}]
          foreach bc $blockCalls {
            :checkCallForTestClassM $bc $featureType $testCaseName "none"
          }
        }
        # handling calls in scenario blocks
        set scens [$c selectNodes scenario]
        foreach s $scens {
          set testScenarioName [$s getAttribute name]
          set scenBlocks [$s selectNodes block]
          foreach b $scenBlocks {
            set blockCalls [$b getAttribute call]
            :aggCalls set [expr {[:aggCalls get] + [llength $blockCalls]}]
            foreach bc $blockCalls {
              :checkCallForTestClassM $bc $featureType $testCaseName $testScenarioName
            }
          }
        }
      }
      puts "all calls [llength [:allCalls get]]"
      puts "aggregatedCalls [:aggCalls get]"

      set aggClassFeatures ""
      foreach cl [[:matrix get] classes get] {
        foreach f [$cl features get] {
          lappend aggClassFeatures [$f name get]
        }
      }

      puts "#ofaggClassFeatures [llength $aggClassFeatures]"
      puts "#ofClasses [llength [[:matrix get] classes get]]"
      puts ------------------
      foreach cl [[:matrix get] testCases get] {
        #puts [$cl name get]
        #puts [llength [$cl features get]]
        #set fs [lsort -unique [$cl features get]]
        #puts [llength [$cl features get]]
        #puts [llength $fs]
        #puts ------------------------------
      }
    }


    :protected method setBasicElements { xmiFile } {
      # set tDOM root element
      set traceXmiData [read [open $xmiFile r]]
      set doc [dom parse $traceXmiData]
      set root [$doc documentElement]

      # set basic elements
      :allCases set [$root selectNodes /TM:TraceModel/testSuite/case]
      foreach ca [:allCases get] {
        set c [TestCase new]
        $c name set [$ca getAttribute name]
        set cs [[:matrix get] testCases get]
        lappend cs $c
        [:matrix get] testCases set $cs

        set currentScens [$ca selectNodes scenario]
        foreach scen $currentScens {
          set sc [TestScenario new]
          $sc name set [$scen getAttribute name]
          set scs [[:matrix get] testScenarios get]
          lappend scs $sc
          [:matrix get] testScenarios set $scs
        }

      }
      :allClasses set [$root selectNodes /TM:TraceModel/class]
      foreach cl [:allClasses get] {
        set c [SutClass new]
        $c name set [$cl getAttribute name]
        set cs [[:matrix get] classes get]
        lappend cs $c
        [:matrix get] classes set $cs
      }
      :allFeatures set [$root selectNodes /TM:TraceModel/feature]
      :allInstances set [$root selectNodes /TM:TraceModel/instance]
      :allCalls set [$root selectNodes /TM:TraceModel/trace/call]
      :aggCalls set 0
    }

    # ...
    # <testScenarioName> (none (in case of case-block call) / <name of current scnario>)
    :protected method checkCallForTestClassM { c featureType testCaseName testScenarioName } {

      set call [lindex [:allCalls get] [string range $c 15 19]]
      set callName [$call getAttribute name]
      set definedByTestScript [$call getAttribute definedBySTF]

      # get the name of the calling feature (caller)
      set callerInd [$call getAttribute caller]
      set caller [lindex [:allFeatures get] [string range $callerInd 11 13]]
      set callerName [$caller getAttribute name]

      # get the calling instance (source)
      set sourceInd [$call getAttribute source]
      set source [lindex [:allInstances get] [string range $sourceInd 12 17]]
      set sourceName [$source getAttribute name]

      # get name of the calling class (providing the calling feature)
      set classSourceInd [$source getAttribute definition]
      set classSource [lindex [:allClasses get] [string range $classSourceInd 9 13]]
      set classSourceName [$classSource getAttribute name]

      # get the name of the called feature (callee)
      set calleeInd [$call getAttribute callee]
      set callee [lindex [:allFeatures get] [string range $calleeInd 11 13]]
      set calleeName [$callee getAttribute name]

      # get the called instance (target)
      set targetInd [$call getAttribute target]
      set target [lindex [:allInstances get] [string range $targetInd 12 17]]
      set targetName [$target getAttribute name]

      # get name of the called class (providing the called feature)
      set classTargetInd [$target getAttribute definition]
      set classTarget [lindex [:allClasses get] [string range $classTargetInd 9 13]]
      set classTargetName [$classTarget getAttribute name]

      ##########
      set currentFeature ""
      switch $featureType {
        called {

          foreach cl [[:matrix get] classes get] {
            if {[$cl name get] eq $classTargetName} {
              # add feature to class's feature list
              set fs [$cl features get]
              set featureSet 0
              foreach f $fs {
                if {[$f name get] eq $calleeName} {
                  set featureSet 1

                  # added
                  if {![$f callDefinedInTestScript get]} {
                    $f callDefinedInTestScript set $definedByTestScript
                  }


                  set currentFeature $f
                  break
                }
              }

              if {!$featureSet} {
                set currentFeature [Feature new]
                $currentFeature name set $calleeName

                # added
                if {![$currentFeature callDefinedInTestScript get]} {
                  $currentFeature callDefinedInTestScript set $definedByTestScript
                }

                lappend fs $currentFeature
                $cl features set $fs
                # set class as feature's class
                $currentFeature class set $cl
                break
              }
            }
          }


          # add feature to case's feature list
          foreach cl [[:matrix get] testCases get] {
            if {[$cl name get] eq $testCaseName} {
              # add feature to case's feature list
              set fs [$cl features get]
              lappend fs $currentFeature
              $cl features set $fs

              #  add case to feature's cases list
              break
            }
          }

          # add feature to scenario's feature list
          if {$testScenarioName ne "none"} {
            foreach sc [[:matrix get] testScenarios get] {
              if {[$sc name get] eq $testScenarioName} {
                set fs [$sc features get]
                lappend fs $currentFeature
                $sc features set $fs
                break
              }
            }

          }

        }

        calling {

          foreach cl [[:matrix get] classes get] {
            if {[$cl name get] eq $classSourceName} {
              # add feature to class's feature list
              set fs [$cl features get]
              set featureSet 0
              foreach f $fs {
                if {[$f name get] eq $callerName} {
                  set featureSet 1
                  set currentFeature $f
                  break
                }
              }

              if {!$featureSet} {
                set currentFeature [Feature new]
                $currentFeature name set $callerName
                lappend fs $currentFeature
                $cl features set $fs
                # set class as feature's class
                $currentFeature class set $cl
                break
              }
            }
          }

          foreach cl [[:matrix get] testCases get] {
            if {[$cl name get] eq $testCaseName} {
              # add feature to case's feature list
              set fs [$cl features get]
              lappend fs $currentFeature
              $cl features set $fs

              #  add case to feature's cases list
              break
            }

          }

        }

        interClassCalling {

          foreach cl [[:matrix get] classes get] {
            if {[$cl name get] eq $classSourceName} {
              # add feature to class's feature list
              set fs [$cl features get]
              set featureSet 0
              foreach f $fs {
                if {[$f name get] eq $callerName} {
                  set featureSet 1
                  set currentFeature $f
                  break
                }
              }

              if {!$featureSet && $classSourceName ne $classTargetName} {
                set currentFeature [Feature new]
                $currentFeature name set $callerName
                lappend fs $currentFeature
                $cl features set $fs
                # set class as feature's class
                $currentFeature class set $cl
                break
              }
            }
          }
          if {$classSourceName ne $classTargetName} {
            foreach cl [[:matrix get] testCases get] {
              if {[$cl name get] eq $testCaseName} {
                # add feature to case's feature list
                set fs [$cl features get]
                lappend fs $currentFeature
                $cl features set $fs

                #  add case to feature's cases list
                break
              }

            }
          }
        }

      }

    }

    # 
    #:public method buildInterClassMatrix { xmiFile testCaseName } {
    #}
    #:public method buildTestToFeatureMatrix { xmiFile testCaseName } {
    #}
    #:public method buildClassToFeatureMatrix { xmiFile testCaseName } {
    #  if {testCaseName eq "allCases"} {
    #  } else {
    #  }
    #}

  }
  nx::Class create NamedElement {
    :property -accessor public {name}
  }

  nx::Class create Matrix {
    :property -accessor public {headLine}
    :property -accessor public {content}


    :public method exportCsv { } {

    }

  }

  nx::Class create TestClassMatrix -superclass Matrix {
    :property -accessor public {testCases ""}
    :property -accessor public {testScenarios ""}
    :property -accessor public {classes ""}

    # testLevel ('case'/'scenario')
    # sutUnit (classes-> 'classes' / class features -> <sutClassName> / features of all classes -> 'allFeatures')
    # scope (all/sut/testScript)
    :public method exportCsv { testLevel sutUnit callScope} {
      set f matrix.csv
      set fileId [open $f "w"]

      set headLine ""
      set csvText ""
      set content ""

      # SUT classes as unit
      if {$sutUnit eq "classes"} {

        #build csvText
        foreach cl [:classes get] {
          append headLine ";[$cl name get]"
        }
        append headLine "\n"

        switch $testLevel {
          case {
            foreach ca [:testCases get] {
              append content "[$ca name get];"

      				#call vs. feature
              set fs [lsort -unique [$ca features get]]
              #set fs [$ca features get]

              foreach cl [:classes get] {
                set numberOfClFeaturesPerCase 0
                foreach f $fs {
                  if {[$cl name get] eq [[$f class get] name get]} {
                    incr numberOfClFeaturesPerCase
                    #append content "[$f name get]  "
                  }
                }
                append content "$numberOfClFeaturesPerCase;"
                #append content ";"
              }
              append content "\n"
            }

          }
          scenario {
            foreach sc [:testScenarios get] {
              append content "[$sc name get];"

      				#call vs. feature
              set fs [lsort -unique [$sc features get]]
              #set fs [$ca features get]

              foreach cl [:classes get] {
                set numberOfClFeaturesPerCase 0
                foreach f $fs {
                  if {[$cl name get] eq [[$f class get] name get]} {
                    incr numberOfClFeaturesPerCase
                    #append content "[$f name get]  "
                  }
                }
                append content "$numberOfClFeaturesPerCase;"
                #append content ";"
              }
              append content "\n"
            }

          }
        }

      # SUT class features as unit -> SUT class name
      } elseif  {$sutUnit eq "allFeatures"} {


        foreach cl [:classes get] {
          foreach f [$cl features get] {
            append headLine ";[$cl name get]::[$f name get]"
          }
        }
        append headLine "\n"

        switch $testLevel {
          case {
          }

          scenario {
            foreach sc [:testScenarios get] {
              append content "[$sc name get];"

      				#call vs. feature
              set fsScenario [lsort -unique [$sc features get]]

              foreach cl [:classes get] {
                foreach f [$cl features get] {

                  set num 0
                  foreach fiS $fsScenario {
                    if { [$cl name get] eq [[$fiS class get] name get] && [$fiS name get] eq [$f name get]} {
                      set num 1
                      break
                    }
                  }
                  append content "$num;"
                }
              }
              append content "\n"

            }

          }

        }


      } else {
        #build csvText
        set selectedClass ""
        foreach cl [:classes get] {
          if {[$cl name get] eq $sutUnit} {
            set selectedClass $cl
            break
          }
        }

        if { $selectedClass eq "" } {
          puts "error in sutUnit"
          exit
        }

        foreach f [$selectedClass features get] {
          append headLine ";[$f name get]"
        }
        append headLine "\n"

        switch $testLevel {
          case {
            foreach ca [:testCases get] {
              append content "[$ca name get];"


      				#call vs. feature
              # !!!! changed!
              #set fsCase [lsort -unique [$ca features get]]
              set fsCase [$ca features get]

              set numberOfClFeaturesCallsPerCase 0
              foreach f [$selectedClass features get] {

                set num 0
                foreach fiC $fsCase {
                  if { [$selectedClass name get] eq [[$fiC class get] name get] && [$fiC name get] eq [$f name get]} {

                    # checking the scope
                    if {$callScope eq "sut"} {
                      if {![$fiC callDefinedInTestScript get]} {
                        set num 1
                      }

                    }

                    # !!!! changed!
                    #set num 1
                    ####incr num
                    break
                  }
                }
                append content "$num;"
              }
              append content "\n"
            }

          }
          scenario {

            foreach sc [:testScenarios get] {
              append content "[$sc name get];"

      				#call vs. feature
              set fsScenario [lsort -unique [$sc features get]]
              #set fsCallsInCase [$ca features get]

              set numberOfClFeaturesCallsPerCase 0
              foreach f [$selectedClass features get] {

                set num 0
                foreach fiS $fsScenario {
                  if { [$selectedClass name get] eq [[$fiS class get] name get] && [$fiS name get] eq [$f name get]} {
                    set num 1
                    break
                  }
                }
                append content "$num;"
              }
              append content "\n"
            }

          }
        }

      }


      append csvText $headLine
      append csvText $content

      #puts $csvText
      puts -nonewline $fileId $csvText
      puts "'matrix.csv' created"
    }

  }
  nx::Class create InterClassMatrix -superclass Matrix {
    :property -accessor public {testCaseName ""}
    :property -accessor public {sourceClasses ""}
    :property -accessor public {targetClasses ""}
  }


  nx::Class create TestCase -superclass NamedElement {
    :property -accessor public {features ""}
  }
  nx::Class create TestScenario -superclass NamedElement {
    :property -accessor public {features ""}
  }

  nx::Class create SutClass -superclass NamedElement {
    :property -accessor public {features ""}
  }

  nx::Class create Feature -superclass NamedElement {
    :property -accessor public {testCases ""}
    :property -accessor public {class ""}
    :property -accessor public {callDefinedInTestScript 0}
  }
}



set file tm_functionalclone1003.xmi
#set file ba-tc01-18-infoClass.xmi

set ta [::TAGG::TraceAggregator new]
# testLevel (below: scenario) not recognized for method
$ta buildTestToSutMatrix $file case feature called all 0
[$ta matrix get] exportCsv scenario allFeatures all
#[$ta matrix get] exportCsv scenario UML2ControlFlow all
#[$ta matrix get] exportCsv case UML2BusinessActivity sut
