
#####################################################################
#
# STORM_i -- scenario-based testing of object-oriented runtime models
# (c) Mark Strembeck
#
#
# transfered from XOTcl to NX 2.0 and extended by tracer methods
# (mixin messageInterceptor and filter DoInterceptor)
# by Thorsten Haendler
#
#####################################################################
package provide STORM_i 0.5.1
#package require nx
#package require TR

namespace eval ::STORM_i {

#filterer messageInterceptor to intercept each unknown feature call
nx::Object method messageInterceptor args {
  set relevant true
  #irrelevant name spaces
  set irrelevantNs { ::xotcl ::nx ::XP ::TM ::TR}
  set irrelevantMthds {"*value*" "*init*" "*object_configureparameter*" "*new*"}
  #define irrelevant calls
  if {[namespace qualifiers [current callingclass]] in $irrelevantNs || [current callingclass] in {"::STORM_i::DoInterceptor" "::STORM_i::MessageInterceptor" ""}
	  || [current callingmethod] in {"messageInterceptor" "assignInvocationContext" ""}  } {
    set relevant false
  }
  foreach m $irrelevantMthds {
	  if {[string match $m [current calledmethod]]} {
		  set relevant false
	  }
	  if {[string match $m [current callingmethod]]} {
			set relevant false
		}
  }
  if $relevant {


		set cs [::TR::CallStack getInstance]
		$cs push S
    set returnValue [next]

    set tracer [::TR::Tracer getInstance]

    set cs [::TR::CallStack getInstance]
		$cs push E
    $cs checkForPop
		#checks
		#puts -----------------------------------------------
    #puts callingMethod:[current callingmethod]
    #puts callingObject:[current callingobject]
    #puts callingClass:[current callingclass]
    #puts calledMethod:[current calledmethod]
    #puts arguments:[current args]
    #puts calledObject:[current object]
    #puts calledClass:[current calledclass]
    #puts rV:$returnValue
    $tracer checkSection [current callingobject] [current callingclass]
    #puts CC:[$tracer currentTestCase]
    #puts CS:[$tracer currentTestScenario]
    #puts CTB:[$tracer currentTestBlock]
		#puts -----------------------------------------------

    #derive the call dependencies of the traced call and add it to the call sequence of trace
    $tracer addTracedCall [current callingmethod] [current callingobject] [current callingclass] [current calledmethod] [current object] [current calledclass] [current args] $returnValue


    return $returnValue
  } else {
    return [next]
  }
}
#mixin DoInterceptor registered by each relevant STF feature
nx::Class create DoInterceptor {
  :public method do {script} {
    set tracer [::TR::Tracer getInstance]
    #checks
    #puts doCallM:[current callingmethod]
    #puts doCallC:[current callingclass]
    #puts doCallO:[current callingobject]
    $tracer currentTestBlock set [current callingmethod]
    #start intercepting every call
    Object filters add messageInterceptor
    #forward the call to the intended method do and store its result
    set result [next]
    #stop intercepting
    Object filters delete messageInterceptor
    #return the result
    return $result
  }
}

#################
# Class TestSuite
#################
nx::Class create TestSuite -mixin DoInterceptor {
	:property -accessor public {name Test1}
  :property -accessor public {intercepted 1}

	:property -accessor public {setup_script ""}
	:property -accessor public {cleanup_script ""}
	:property -accessor public {detailed_report 0}
	:property -accessor public {halt_on_first_error 0}
	:property -accessor public {order ""}
	:property -accessor public {tr ""}

	:public method init { } {
    #assign scenarios to testcases
    :assignScenariosToCases

    #start reporting
		:tr set [::STORM_i::Report new]
		[:tr get] addLine "----------------------------------------------"
		[:tr get] addLine "Begin Test Run: [clock format [clock seconds]]"
		[:tr get] addLine "----------------------------------------------\n"
		if {[:detailed_report get]} {[:tr get] detailed set 1}
	}

  #check all instances of scenarios and testcases and
  #assign scenarios to related testcases
  :public method assignScenariosToCases {} {
    set cases [::STORM_i::TestCase info instances]
    set scenarios [::STORM_i::TestScenario info instances]
    foreach c $cases {
      foreach s $scenarios {
        if {[$c name get] eq [$s testcase get]} {
          $c addScenario $s
        }
      }
    }
  }
	:public method do {script} {
		eval $script
	}
  #run a single testcase
	:public method runTestCase {tc} {
    set returnVal 0

    $tc setup
    #check preconditions
		set pre [$tc checkPreConditions]
		if {![$pre success get]} {
			[:tr get] log -TestCase [$tc name get] "FAILED" "TestCase [$tc name get] incomplete -> [$pre comment get]"
      $tc cleanup
			return $returnVal
		}
    #run testscenarios
    #testcase's setup and cleanup can not be done here:
		set scenarios [$tc runTestScenarios]

		if {![$scenarios success get]} {
			[:tr get] log -TestCase [$tc name get] "FAILED" "TestCase [$tc name get] incomplete\n -> Failed test scenario:\
					[[$scenarios object_reference get] name get]\n - Message: [$scenarios comment get]"
      $tc cleanup
			return $returnVal
		}
    #check postconditions
		set post [$tc checkPostConditions]
		if {![$post success get]} {
			[:tr get] log -TestCase [$tc name get] "FAILED" "TestCase [$tc name get] incomplete - [$post comment get]"
      $tc cleanup
			return $returnVal
		}

    #in case that every part of testcase is successful
    set msg ""
    foreach s [$tc order get] {
      append msg "\n   -TestScenario [$s name get] completed"
    }
		[:tr get] log "\n-TestCase" [$tc name get] "Successful" $msg;#modified: "TestCase [$tc name get] completed"


    $tc cleanup
    set returnVal 1
    return $returnVal
	}

	#run all testcases
	:public method runAllTestCases { } {
		set err 0
		if {[:order get] == ""} {
			:order set [::STORM_i::TestCase info instances]

		}

    #do testsuite's setup
		if {[:setup_script get] ne ""} { :do [:setup_script get] }

    #run each testcase
		foreach t [:order get] {
			if {![:runTestCase $t]} {
				set err 1
				if {[:halt_on_first_error get]} {

					if {[:cleanup_script get] ne ""} { :do [$t cleanup_script get] }
						return 0
				}
				continue
			}
		}

    #do testsuite's cleanup
		if {[:cleanup_script get] ne ""} { :do [:cleanup_script get] }


    ##
    #after executing the test run, order the calls,
    # create a config model and export trace and config
    # model as xmi files
    ##########################################################################
    set tracer [::TR::Tracer getInstance]
    #order calls


    set cs [::TR::CallStack getInstance]
    set indizes [$cs indizes get]
    puts ------------indizes---------------------------------------
    puts l_indizes:__[llength $indizes]
    puts indizes:__$indizes

    set tracer [::TR::Tracer getInstance]
    set traceOrder [[[$tracer traceModel get] trace get] calls get]
    puts l_traceOrder:__[llength $traceOrder]
    puts ------------old---------------------------------------
    foreach c $traceOrder {
      puts [$c name get]__[[[$c source get] definition get] name get]__[[[$c target get] definition get] name get]
    }
    puts pairs-----------
    for {set i 0} {$i < [llength $indizes]} {incr i} {
    #foreach i $indizes {}
      puts [lindex $indizes $i]__[lindex $traceOrder $i]
      set pair "[lindex $indizes $i] [lindex $traceOrder $i]"
      lappend indexedCalls $pair
    }
    #foreach i $indexedCalls {}
    for {set i 0} {$i < [llength $indexedCalls]} {incr i} {
      if { [expr {[llength [lindex $indexedCalls $i]] == 0 }]} {
        puts index__$i
        #lreplace $indexedCalls $i $i hallo
        # break
      } else {
        lappend indexedCalls2 [lindex $indexedCalls $i]
      }

    }

    #puts indexedCalls2_$indexedCalls2
    set sortedIndexedCalls [lsort -integer -index 0 $indexedCalls2]
    #puts --------------------------------
    #puts sorted:_____$sortedIndexedCalls

    foreach comb $sortedIndexedCalls {
      lappend orderedCalls [lindex $comb 1]
    }
    puts l_orderedCalls:__[llength $orderedCalls]
    puts ------------new---------------------------------------
    foreach c $orderedCalls {
      puts [$c name get]__[[[$c source get] definition get] name get]__[[[$c target get] definition get] name get]
    }

    [[$tracer traceModel get] trace get] calls set $orderedCalls
    # create dictionairy (keys = indizes, values = tarceOrder)
    #for {set i 0} {$i < [llength $indizes]} {incr i} {
	  #   dict set indexCallDict [lindex $indizes $i] [lindex $traceOrder $i]
    #}
    #puts $indexCallDict
    #puts values___[dict values $indexCallDict]
    #set sortedKeys [lsort -integer [dict keys $indexCallDict]]
    #puts sorted____$sortedKeys



    #set ttt [lsort {1,2,4,5,12,15,21}]
    #puts ttt__$ttt
    #[[$tracer traceModel get] trace get] calls set [$tracer orderCalls [[[$tracer traceModel get] trace get] calls get]]
    ###########################################################################


  #check
	#foreach call [[[$tracer traceModel get] trace get] calls get] {
	#	puts [[$call caller get] name get]__calls__[[$call callee get] name get]__defBySTF:[$call definedBySTF get]__isCon:[[$call callee get] isConstructor get]__block:[[$call testBlock get] name get]__scenario:[[[$call testBlock get] scenario get] name get]__case:[[[$call testBlock get] case get] name get]

	#}
    #instance of trace model xmi exporter
    set xp [::XP::XMIExporter new -traceModel [$tracer traceModel get] -nameOfTraceModel [:name get]]
    $xp createXMITraceModel
    ##


	[:tr get] addLine \n

		if {!$err} {
			#puts normalRun
		      [:tr get] log ->TestSuite [:name get] "Successful" "TestSuite [:name get] completed\n"
		      [:tr get] addLine "Successfully tested [:testCaseCount] test cases including\
        			   [:testScenarioCount] test scenarios"
        	      [:tr get] addLine "The [:testCaseCount] test cases contain [:testCasePreconditionCount]\
        			   precondition tests and [:testCasePostconditionCount] postcondition tests."
        	      [:tr get] addLine "The [:testScenarioCount] test scenarios contain [:testScenarioPreconditionCount]\
        			   precondition tests and [:testScenarioPostconditionCount] postcondition tests."
        	      [:tr get] addLine "\n----------------------------------------------"
        	      [:tr get] addLine "End Test Run: [clock format [clock seconds]]"
        	      [:tr get] addLine "----------------------------------------------"
								[:tr get] returnResult
								#return 1
        	} else {
        	      [:tr get] log ->TestSuite [:name get] "FAILED" "TestSuite [:name get] incomplete"
        	      [:tr get] addLine "\n----------------------------------------------"
        	      [:tr get] addLine "End Test Run: [clock format [clock seconds]]"
        	      [:tr get] addLine "----------------------------------------------"
								[:tr get] returnResult
        	      #return 0
        	}
	  }

	  :public method testCaseCount {} {
	  	if {[:order get] != ""} {
	  		return [llength [:order get]]
	  	}
	  	return [llength [::STORM_i::TestCase info instances]]
	  }

	  :public method testScenarioCount {} {
	  	if {[:order get] == ""} {
	  		:order set [::STORM_i::TestCase info instances]
	  	}

	  	set count 0
	  	foreach tc [:order get] {
	  		if {[$tc order get] != ""} {
	  			incr count [llength [$tc order get]]
	  		} else {
	  			incr count [llength [$tc info children]]
	  		}
	  	}
	  	return $count
	  }

	  :public method testCasePreconditionCount {} {
	  	if {[:order get] == ""} {
	  		:order set [::STORM_i::TestCase info instances]
	  	}
	  	set count 0
	  	foreach tc [:order get] {
	  		incr count [$tc preconditionCount]
	  	}
	  	return $count
	  }

	 :public method testCasePostconditionCount {} {
		if {[:order get] == ""} {
			:order set [::STORM_i::TestCase info instances]
		}
		set count 0
		foreach tc [:order get] {
			incr count [$tc postconditionCount]
		}
		return $count
	 }

	 :public method testScenarioPreconditionCount {} {
	 	if {[:order get] == ""} {
	 		:order set [::STORM_i::TestCase info instances]
	 	}
	 	set count 0
	 	foreach tc [:order get] {
	 		incr count [$tc testScenarioPreconditionCount]
	 	}
	 	return $count
	 }

	 :public method testScenarioPostconditionCount {} {
	 	if {[:order get] == ""} {
	 		:order set [::STORM_i::TestCase info instances]
	 	}
	 	set count 0
	 	foreach tc [:order get] {
	 		incr count [$tc testScenarioPostconditionCount]
	 	}
	 	return $count
	 }

	 :public method destroy args {
		[:tr get] print
		[:tr get] returnResult
	 	next
	 }
}
################
# Class TestCase
################
nx::Class create TestCase -mixin DoInterceptor {
	:property -accessor public {setup_script ""}
	:property -accessor public {preconditions ""}
	:property -accessor public {postconditions ""}
	:property -accessor public {cleanup_script ""}
	:property -accessor public {order ""}
	:property -accessor public {name}

	:public method do {script} {
		eval $script
	}
  #add a scenario to the testcase's order
	:public method addScenario {s} {
		set scens [:order get]
		lappend scens $s
		:order set $scens
	}

  :public method setup {} {
    if {[:setup_script get] ne ""} { :do [:setup_script get] }
  }

  #check preconditions
	:public method checkPreConditions {} {

		foreach pc [:preconditions get] {
			if {![:do $pc] } {
				return [::STORM_i::TestResult new\
					-success 0 -object_reference [self]\
					-comment "precondition << $pc >> failed"]
			}
		}
		return [::STORM_i::TestResult new -success 1 -object_reference [self]]
	}

	:public method checkPostConditions {} {
		foreach pc [:postconditions get] {
			if {![:do $pc]} {
				return [::STORM_i::TestResult new\
					-success 0 -object_reference [self]\
					-comment "postcondition << $pc >> failed"]
			}
		}
		return [::STORM_i::TestResult new\
			-success 1 -object_reference [self]]
	}

  :public method cleanup {} {
    if {[:cleanup_script get] ne ""} { :do [:cleanup_script get] }

  }


	:public method runTestScenarios {} {

    ##do testcase's setup (previous place in xotcl version)
    #if {[:setup_script] ne ""} { :do [:setup_script] }

    ##not necessary, does not work after modification of scenario assignment
		#if {[:order] == ""} {
		#	:order [lsort -dictionary [:info children]]
		#}

    #each scenario assigned to testcase
		foreach s [:order get] {
			if {[$s info has type ::STORM_i::TestScenario]} {

        #do setup
        $s setup

        set pre [$s checkPreConditions]
				if {![$pre success get]} {
					if {[:cleanup_script get] ne ""} { :do [:cleanup_script get] }
					return [::STORM_i::TestResult new\
						-success 0 -object_reference $s\
						-comment "Test scenario [$s name get] -> [$pre comment get]\n"]
				}

				set testbody [$s test]
				if {![$testbody success get]} {
					$s cleanup
					if {[:cleanup_script get] ne ""} { :do [:cleanup_script get] }
					return [::STORM_i::TestResult new\
						-success 0 -object_reference $s\
						-comment "Test scenario [$s name get] -> [$testbody comment get]\n"]
				}
				set post [$s checkPostConditions]
				if {![$post success get]} {
					$s cleanup
					if {[:cleanup_script get] ne ""} { :do [:cleanup_script get] }
					return [::STORM_i::TestResult new\
						-success 0 -object_reference $s\
						-comment "Test scenario [$s name get] -> [$post comment get]\n"]
				}
        #do cleanup
				$s cleanup
			}
		}
    ##do testcase's cleanup (previous place in xotcl version)
		#if {[:cleanup_script] ne ""} { :do [:cleanup_script] }
		return [::STORM_i::TestResult new\
			-success 1 -object_reference [self]\
			-comment "Test case :successful"]
	  }
	:public method preconditionCount {} {
		 return [llength [:preconditions get]]
	}

	:public method postconditionCount {} {
	 	return [llength [:postconditions get]]
	}

	:public method testScenarioPreconditionCount {} {
	      	set count 0
	      	if {[:order get] == ""} {
	      		:order set [lsort -dictionary [:info children]]
	      	}
	      	foreach s [:order get] {
	      		incr count [$s preconditionCount]
	      	}
	      	return $count
	}

	:public method testScenarioPostconditionCount {} {
	      	set count 0
	      	if {[:order get] == ""} {
	      		:order set [lsort -dictionary [:info children]]
	      	}
	      	foreach s [:order get] {
	      		incr count [$s postconditionCount]
	      	}
	      	return $count
	}

	:public method resetAllAutonames {class} {
	      	$class array unset __autonames
	}
}

####################
# Class TestScenario
####################

nx::Class create TestScenario -mixin DoInterceptor {

	:property -accessor public {setup_script ""}
	:property -accessor public {test_body ""}
	:property -accessor public {expected_result ""}
	:property -accessor public {preconditions ""}
	:property -accessor public {postconditions ""}
	:property -accessor public {cleanup_script ""}
	:property -accessor public {name}
	:property -accessor public {testcase}

	:public method init args {
		next
	}

	# from tcllib ::struct::list:Lequal
	:public method lequal {a b} {
		if {[llength $a] != [llength $b]} {
			return 0
		}
		if {[lindex $a 0] == $a} {
			return [string equal $a $b]
		}
		foreach i $a j $b {
			if {![:lequal $i $j]} { return 0 }
		}
		return 1
	}

  #
	:public method do {script} {
		eval $script
	}

  #added:
  #do setup
  :public method setup {} {
      if {[:setup_script get] ne ""} {
        :do [:setup_script get]
      }
   }

  #check preconditions
	:public method checkPreConditions {} {
		foreach pc [:preconditions get] {
			if {![:do $pc]} {
				return [::STORM_i::TestResult new\
					-success 0 -object_reference [self]\
					-comment "precondition << $pc >> failed"]
			}
		}
		return [::STORM_i::TestResult new\
			-success 1 -object_reference [self]]
	}

  #execute testbody
  #compare expected result and result, return success
  :public method test {} {
    if {[:test_body get] ne ""} {
      set result [:do [:test_body get]]

      if {[:expected_result get] ne ""} {
          #TODO transform to method with arguments expected_result and result
          set resultequal [:lequal $result [:expected_result get]]
          if {$resultequal} {
            return [::STORM_i::TestResult new -success 1\
              -object_reference [self]]
          } else {
            return [::STORM_i::TestResult new -success 0\
              -object_reference [self]\
              -comment "[self]: test body result << $result >> does not match the\
              expected result << [:expected_result get] >>."]
          }
        } else {
          return [::STORM_i::TestResult new\
            -success 1 -object_reference [self]\
            -comment "[self]: no expected result defined"]
        }
      } else {
        return [::STORM_i::TestResult new\
          -success 0 -object_reference [self]\
          -comment "[self]: no test body defined"]
      }
     }

  #check postconditions
	:public method checkPostConditions {} {
		foreach pc [:postconditions get] {
			if {![:do $pc]} {
				return [::STORM_i::TestResult new\
					-success 0 -object_reference [self]\
					-comment "postcondition << $pc >> failed"]
			}
		}
		return [::STORM_i::TestResult new\
			-success 1 -object_reference [self]]
	}

  #do cleanup
	:public method cleanup {} {
		if {[:cleanup_script get] ne ""} {
			:do [:cleanup_script get]
		}
	}


	 :public method preconditionCount {} {
	  	return [llength [:preconditions get]]
	 }

	 :public method postconditionCount {} {
	  	return [llength [:postconditions get]]
	 }

}

##############
# Class Report
##############
nx::Class create Report {
	:property -accessor public {detailed 0}
	:property -accessor public {report ""}
	:public method addLine {l} {
		set lines [:report get]
    append lines "$l \n"
    :report set $lines


	}

	:public method log {type test status message} {
		if {![:detailed get]} {
			:report set [append x [:report get] "$type $test **$status** \n"]

		} else {
			:report set [append x [:report get] "$type $test **$status** message: $message \n"]
		}
	}

	:public method print {{channel stderr}} {

		puts $channel "\n++++++ Begin TEST REPORT ++++++ \n"
		puts $channel [:report get]
		puts $channel "\n++++++ End TEST REPORT ++++++++ \n"

	}

	:public method returnResult {} {

		set result ""
		append result "\n++++++ Begin TEST REPORT ++++++ \n"
		append result [:report get]
		append result "\n++++++ End TEST REPORT ++++++++ \n"
		return $result
	}


	:public method save {file} {

		set output [open $file a+]
		puts $output "\n++++++ Begin TEST REPORT ++++++ \n"
		puts $output [:report]
		puts $output "\n++++++ End TEST REPORT ++++++ \n"
		close $output
	}
}
##################
# Class TestResult
##################
nx::Class create TestResult {
	:property -accessor public {success {} }
	:property -accessor public {object_name {} }
	:property -accessor public {object_reference {} }
	:property -accessor public {comment {} }

}

}; #closing namespace
