package provide TM 0.1
package require nx 
#####################################################################
#
# KaleidoScope - Trace Model
# (c) Thorsten Haendler
#
#####################################################################
namespace eval ::TM {
	nx::Class create TraceModel {
	  :property -accessor public {name traceModel1}
	  #one instance of class Trace contains call sequence
	  :property -accessor public {trace}
	  #instances of Class and Instance are stored directly by the trace model
	  :property -accessor public {class {} }
	  :property -accessor public {instance {} }
	  :property -accessor public {feature {} }
	  :property -accessor public {testSuite ""}
		:property -accessor public {component {} }
	  #storage of current id
	  :property -accessor public {indID 0}
	  #provide individual id for each element
	  :public method useID {} {
	    :indID set [expr {[:indID] +1}]
	    return [:indID get] 
	  }
	  #singleton of class TestModel
	  :public object method getInstance {} {
	    if {[llength [:info instances]]  == 0} {
	      set tm [TraceModel new] 
	      return $tm  
	    } elseif {[llength [:info instances]]  == 1} {
	      return [:info instances]
	    } 
	  }
	} 
	
	#abstract class NamedElement with name and id
	nx::Class create NamedElement {
	  :property -accessor public {name}
	  :property -accessor public {id}  
	} 
	# class Trace aggregates all calls in an ordered sequence
	nx::Class create Trace -superclass NamedElement {
	  :property -accessor public {calls {} }
	  
	  :public method addCall {call} {
	      #set id 13;#[:useID] 
	      #$call id set [$call name get]$id
	      
	      set cs [:calls get]
	      lappend cs $call
	      :calls set $cs
	    }
	}  
	
	# class Call with associations to Instance (target, source), to Feature (caller, callee),
	# and to Literal (arguments, return value)
	nx::Class create FeatureCall -superclass NamedElement {
	  :property -accessor public {target}
	  :property -accessor public {source}
	  :property -accessor public {testBlock}
	  :property -accessor public {returnValue ""} 
	  :property -accessor public {caller ""}
	  :property -accessor public {callee ""}
	  :property -accessor public {argument ""}
	  :property -accessor public {returnVal ""}
	  :property -accessor public {definedBySTF false}
	  
	  :public method addArgument {a} {
	      set args [:argument get]
	      lappend args $a
	      :argument set $args   
	    }
	}   

	# class Argument
	nx::Class create Argument -superclass NamedElement {
	  :property -accessor public {value}
	}  
	# class ReturnValue
	nx::Class create ReturnValue -superclass NamedElement {
	  :property -accessor public {value}	
	}  
	# class Feature 
	nx::Class create Feature -superclass NamedElement {
	  :property -accessor public {isConstructor false}
	  :property -accessor public {isDestructor false}  
	  :property -accessor public {owningClass}
	}  
	# class Instance 
	nx::Class create Instance -superclass NamedElement {
	  :property -accessor public {definition}  
	}  
	# class Class  
	nx::Class create Class -superclass NamedElement { 
	  :property -accessor public {ownedFeature ""}
	}  
	# class TestSuite containing cases and blocks
	nx::Class create TestSuite -superclass NamedElement {
	  :property -accessor public {case ""}
	  :property -accessor public {block ""}
	}
	# class TestCase containing scenarios and blocks
	nx::Class create TestCase -superclass NamedElement {  
	  :property -accessor public {scenario ""}  
	  :property -accessor public {block ""}	
	}
	# class TestScenario containing blocks
	nx::Class create TestScenario -superclass NamedElement {  
	  :property -accessor public {block ""}
	  #:property -accessor public {case ""}
	}  
	# class TestBlock with related calls
	nx::Class create TestBlock -superclass NamedElement {
	  :property -accessor public {call ""}
	  :property -accessor public {scenario ""}
	  :property -accessor public {case ""}	
	}	
  # class Component
  nx::Class create Component -superclass NamedElement {
		:property -accessor public {class ""}
		
		#check if a class is already registered for a component
  	# returns the - in any case - registered class
		:public method addClass {cl} {
			set ex 0
    	foreach c [:class get] {
				
      	if {[$c name get] eq [$cl name get]} {
        	set ex 1  
      	} 
    	}

    	if {!$ex} {
   
      	set cls [:class get]
      	lappend cls $cl
      	:class set $cls    

    	}
    	
  	}

	}


}
