package require nx 
package provide VM 0.1
package require tdom

#####################################################################
#
# KaleidoScope - View-Model Creator
# (c) Thorsten Haendler
#
#####################################################################
namespace eval ::VM {

  nx::Class create VMCreator {
    :property -accessor public {path:substdefault [pwd]/files/tmp} ;#/xmiFiles
		
		# configure test partition and call scope of a view model:
		# vName: name of the resulting view model and XML file 
		# callScope: "stfToSut", "SutIntern", or "both" (StfToSut and SutIntern) 
		# testCase: <name> or ""
		# testScenario: <name> or ""
		# testBlock: <name> or ""    
		# isEntireTS: boolean ("true, if no specific test partition is selected)
		:public method createXMIModel {vName callScope testCase testScenario block isEntireTS} {
      set partition [::VM::TestPartition new -name $vName -testCaseName $testCase -testScenarioName $testScenario -testBlockName $block -isEntireTS $isEntireTS]
      set view [::VM::View new -name $vName -callScope $callScope -partition $partition]
      
      if {[:checkView $view]} {
		    set xmiText [:buildViewModel $view]
		    #set xmiText [:buildHeader]
		    #append xmiText [:buildView $view]
		    set path [file join [:path get] [$view name get].xmi]
		    set f $path
		    set fileId [open $f "w"]
		    puts -nonewline $fileId $xmiText
		    close $fileId
		    puts "XMI view model [$view name get] can be found at $path"        
      } else { 

        #if {[file exists [$view name get].xmi]} {file delete [$view name get].xmi}
        puts "ERROR: invalid test block" 
        exit
        }
    } 	

  # build the XMI view model with tdom
  :public method buildViewModel { view } {
    
    set doc [dom createDocument VM:View]
    set v [$doc documentElement]

    $v setAttribute name [$view name get]
    $v setAttribute xmi:version "2.0"
    $v setAttribute xmlns:xmi "http://www.omg.org/XMI"
    $v setAttribute xmlns:xsi "http://www.w3.org/2001/XMLSchema-instance"
    $v setAttribute xmlns:VM "http://view/1.0"
    $v setAttribute xsi:schemaLocation "http://view/1.0 VIEW.ecore"
    
    $v setAttribute callScope [$view callScope get]
    
    set p [$doc createElement partition] 
    $p setAttribute name [[$view partition get] name get]
    $p setAttribute isEntireTestSuite [[$view partition get] isEntireTS get]
    if {[[$view partition get] testBlockName get] ne ""} { $p setAttribute testBlock [[$view partition get] testBlockName get]}
    if {[[$view partition get] testScenarioName get] ne ""} { $p setAttribute testScenario [[$view partition get] testScenarioName get]}   
    if {[[$view partition get] testCaseName get] ne ""} { $p setAttribute testCase [[$view partition get] testCaseName get]}    
    
    $v appendChild $p

    return [$v asXML]
  }
  
  # check whether the view is valid
  :public method checkView { view } {
    # check the configured test block 
    if {[[$view partition get] testBlockName get] in {"" "all" "null" "setup" "checkPreConditions" "test" "checkPostConditions" "cleanup"} } {
      # also check the configured view regarding the actual trace model TODO
      return 1
    } else { return 0 }
  }

} 	

  nx::Class create View { 	
    :property -accessor public {name}
    :property -accessor public {callScope}
    :property -accessor public {partition}
  }
  nx::Class create TestPartition { 
    :property -accessor public {name}
    :property -accessor public {testCaseName}
    :property -accessor public {testScenarioName}
    :property -accessor public {testBlockName}
    :property -accessor public {isEntireTS}
  }

  
}

#set vmc [::VM::VMCreator new]
#$vmc createXMIModel vXXX stfToSut "" "pushOnFullStack" "" false


